import argparse

from erin import app
from erin.cli import CommandFactory


class StartCommand(CommandFactory):
    def __init__(self, parser, *args, **kwargs):
        self.parser = parser.add_parser(*args, **kwargs)
        self.parser.set_defaults(action=self.run)

        # Config File
        self.parser.add_argument(
            "--config",
            type=argparse.FileType("r"),
            help="give a path to custom config file",
        )

        # Set Logging Levels
        self.choices = [
            "spam",
            "debug",
            "verbose",
            "info",
            "notice",
            "warning",
            "success",
            "error",
            "critical",
        ]
        self.parser.add_argument(
            "--log", choices=self.choices, type=str.lower, help="amount of info to log"
        )

    def run(self, *sys_args, **kwargs):
        passed_args = {"config_file": None, "log_level": None}
        if sys_args[0].config:
            passed_args["config_file"] = sys_args[0].config
        if sys_args[0].log:
            passed_args["log_level"] = sys_args[0].log

        app.start(**passed_args)



import importlib
import logging
import os
import pkgutil
import sys
from pathlib import Path
from typing import List

from erin.core.exceptions import EnvironmentVariableError, PluginNotFoundError

logger = logging.getLogger(__name__)


def find_plugins(package) -> List[str]:
    """
    Finds all top level subpackages in a package and presents them in
    the format required by :meth:`discord.ext.cli.Bot.load_extension`.

    This is useful when you need to load cogs from multiple
    areas of your bot. Simply convert your cogs directory
    into a package and run this method on it.

    Parameters
    -----------
    package : package
        Your package as a python package or a path to one.
        Note: All packages are modules, all modules are not packages.

    Returns
    --------
    list or None
        A list of strings of format `foo.bar` as required by
        :meth:`discord.ext.cli.Bot.load_extension`. If package passed is not
        valid then `None` is returned instead.
    """
    # Check if parameter is a package
    if hasattr(package, "__path__"):
        plugins = pkgutil.walk_packages(package.__path__)
        package_name = os.path.basename(package.__path__[0])
    elif isinstance(package, str):
        if os.path.exists(package):
            plugins = pkgutil.walk_packages([package])
            package_name = os.path.basename(package)
        else:
            raise PluginNotFoundError(package)
    elif isinstance(package, Path):
        if package.exists():
            plugins = pkgutil.walk_packages([package])
            package_name = os.path.basename(str(package))
        else:
            raise PluginNotFoundError(package)
    else:
        raise TypeError(
            f"expected package, str, pathlib.Path or os.PathLike "
            f"object, not {type(package).__name__}"
        )

    # Create plugin import list
    plugins = [f"{package_name}.{i.name}" for i in plugins]
    return plugins


def get_plugin_data(plugin):
    """
    Retrieve the plugin_data dictionary defined in a plugin.
    Parameters
    -----------
    plugin : path to a plugin in module import format
    Returns
    --------
    dict or None
        The plugin_data dict defined in a plugin or None if the dict is
        not defined.
    """
    logger.debug(f"Attempting Plugin Import: {plugin}")
    plugin = importlib.import_module(plugin)
    plugin_data = None
    try:
        plugin_data = plugin.plugin_data
    except AttributeError:
        plugin_data = None
    finally:
        del sys.modules[plugin.__name__]
        return plugin_data


def config_loader(mappings, optional_envs):
    for category, settings in mappings.items():
        for setting, value in settings.items():
            if value.startswith("ERIN_"):
                if value in os.environ:
                    mappings[category][setting] = os.environ[value]
                elif value in optional_envs:
                    mappings[category][setting] = None
                else:
                    raise EnvironmentVariableError(
                        f"{value} is not optional.\n"
                        "Set this in your TOML config file or use 'export "
                        f"{value}=YOUR_CUSTOM_VALUE'!"
                    )
    return mappings



import re


def get_command_args(ctx, lower_case=True):
    """
    Gets the arguments passed to a command.

    :param ctx: pass a :class:`discord.ext.cli.Context` object
    :param lower_case: returns arguments in lower case
    :return: :obj:`list`
    """
    args = ctx.message.content.split(" ")
    if lower_case:
        args = [item.strip().lower() for item in args][1:]
    else:
        args = [item.strip() for item in args][1:]
    return args


def find_members(ctx):
    """
    Parses arguments passed to a command and returns a list of me

    :param ctx: pass a :class:`discord.ext.cli.Context` object
    :return: a list of :class:`discord.Member` objects
    """
    args = get_command_args(ctx)
    members = []
    for arg in args:
        arg = re.findall(r"\d{18}", arg)[0]
        if len(arg) == 18:
            member_obj = ctx.guild.get_member(int(arg))
            if member_obj:
                members.append(member_obj)
    members = list(set(members))
    return members



import os
from copy import deepcopy
from random import random

import pytest

from erin.core.exceptions import EnvironmentVariableError
from erin.core.schema import ENV_MAPPINGS, OPTIONAL_ENVS
from erin.core.utils import config_loader


def test_config_env_vars():
    # Delete environment variables
    delete_envs()

    for key, val in ENV_MAPPINGS.items():
        for sub_key, sub_val in val.items():
            # Some random strings
            if sub_val in OPTIONAL_ENVS:
                os.environ[sub_val] = str(random())[2:]
    with pytest.raises(EnvironmentVariableError) as e_info:
        mappings = config_loader(ENV_MAPPINGS, OPTIONAL_ENVS)

    # Check if ENV variables get saved
    saved_maps = deepcopy(ENV_MAPPINGS)
    for key, val in ENV_MAPPINGS.items():
        for sub_key, sub_val in val.items():
            saved_maps[key][sub_key] = str(random())[2:]
            os.environ[sub_val] = saved_maps[key][sub_key]
    mappings = config_loader(ENV_MAPPINGS, OPTIONAL_ENVS)

    assert mappings == saved_maps

    # Teardown
    delete_envs()


def delete_envs():
    for key, val in ENV_MAPPINGS.items():
        for sub_key, sub_val in val.items():
            try:
                # Some random strings
                del os.environ[sub_val]
            except KeyError:
                pass



# Package Info
import logging
import sys

from verboselogs import VerboseLogger

dist_name = "Erin"
__version__ = "0.1.0-alpha.2"

# Constants
LOG_FORMAT = "%(asctime)s %(name)-18s %(levelname)-8s %(message)s"
DATE_FORMAT = "[%Y-%m-%d %H:%M:%S %z]"

# Set Library Logging Formats
logging.setLoggerClass(VerboseLogger)
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(logging.Formatter(LOG_FORMAT, DATE_FORMAT))
root_logger = logging.getLogger(__name__)
root_logger.addHandler(stream_handler)



class CoreError(Exception):
    """Base exception class for Core modules and internal use"""

    pass


class UserError(Exception):
    """Base exception class raised as a byproduct of misconfiguration"""

    pass


class PluginError(Exception):
    """Base exception class for plugin errors"""

    pass


class PluginNotFoundError(PluginError, FileNotFoundError):
    """
    Raised when a plugin path is requested but doesn’t exist.
    Corresponds to errno `ENOENT`.
    """

    def __init__(self, path, message=None):
        if not message:
            message = f"[Errno 2] No such path: '{path}'"
        super().__init__(message)


class EnvironmentVariableError(UserError):
    """
    Error raised when an environment variable is not configured
    properly.
    """

    def __init__(self, message):
        super().__init__(message)


class DatabaseError(CoreError):
    """
    Error raised when a database operation fails.
    """

    pass


class DatabaseKeyError(DatabaseError):
    """
    Raised when an attempt to fetch from the key value store fails.
    """

    def __init__(self, message):
        super().__init__(message)


class PrimaryKeyError(DatabaseError):
    """
    Raised when there is either a primary key is missing or required.
    """

    def __init__(self, message):
        super().__init__(message)


class TableNotFoundError(DatabaseError):
    """
    Raised when a table is missing.
    """

    def __init__(self, message):
        super().__init__(message)


class RecordExistsError(DatabaseError):
    """
    Raised when a record already exists.
    """

    def __init__(self, message):
        super().__init__(message)


class DatabaseTypeError(DatabaseError):
    """
    Raised when parameters passed are incorrect.
    """

    def __init__(self, message):
        super().__init__(message)



from pathlib import Path

import pytest
from schema import SchemaError

from erin.core.exceptions import PluginNotFoundError
from erin.core.schema import plugin_schema
from erin.core.utils import find_plugins, get_plugin_data
from tests import fake_plugins


def test_find_plugins():
    """
    These tests ensure that the :meth:erin.core.utils.find_plugins`
    used for retrieving import paths is working correctly.
    """
    # Test types of paths
    plugins_list = ["fake_plugins.broken", "fake_plugins.core",
                    "fake_plugins.schema"]
    assert hasattr(fake_plugins, "__path__")
    assert find_plugins(fake_plugins) == plugins_list
    assert find_plugins("tests/fake_plugins") == plugins_list
    assert find_plugins(Path("tests/fake_plugins")) == plugins_list

    # Test exceptions
    with pytest.raises(PluginNotFoundError) as e_info:
        find_plugins("tests/incorrect_plugins_path")

    with pytest.raises(PluginNotFoundError) as e_info:
        find_plugins(Path("tests/incorrect_plugins_path"))

    with pytest.raises(TypeError) as e_info:
        find_plugins(123)


def test_get_plugin_data():
    """
    This is used to test if plugin data is imported correctly.
    """
    plugins_data_dict = {
        "tests.fake_plugins.broken": None,
        "tests.fake_plugins.core": {
            "name": "Test Core Plugins"
        },
        "tests.fake_plugins.schema": {
            "name": "Schema Plugin",
            "database": "enabled"
        },
    }

    for plugin, data in plugins_data_dict.items():
        plugin_data = get_plugin_data(plugin)
        if data:
            assert plugin_data["name"] == data["name"]
        else:
            assert plugin_data == data

        if plugin == "tests.fake_plugins.schema":
            # Ensure Schema Works
            with pytest.raises(SchemaError) as e_info:
                plugin_schema.validate(plugin_data)



import asyncio
import logging
import os
import sys
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

import toml
from discord_slash import SlashCommand

import erin
from erin.client import ErinClient
from erin.core.schema import ENV_MAPPINGS, OPTIONAL_ENVS, config_schema
from erin.core.utils import config_loader

logger = logging.getLogger(__name__)
root_logger = logger.parent

# Set Discord Logging Formats
discord_logger = logging.getLogger("discord")
discord_logger.setLevel(logging.INFO)

# Global Bot Variable
bot = None


def start(**kwargs):
    """
    Starts the bot and obtains all necessary config data.
    """
    if kwargs["log_level"]:
        # Set logger level
        level = logging.getLevelName(kwargs["log_level"].upper())
        root_logger.setLevel(level)
    else:
        root_logger.setLevel("INFO")
    # Config Loader
    try:
        if kwargs["config_file"]:
            config = toml.load(kwargs["config_file"])
        else:
            config = toml.load("erin/erin.toml")
    except FileNotFoundError:
        logger.notice(
            "No config file provided. " "Checking for environment variables instead."
        )
        config = config_loader(ENV_MAPPINGS, OPTIONAL_ENVS)

    # Override configs from config file with ones from cli
    if kwargs["log_level"]:
        config["bot"]["log_level"] = kwargs["log_level"].upper()

    # Validate Config
    config_schema.validate(config)

    logger.info(f"Starting Erin: {erin.__version__}")

    # Discord Debug Logging
    if config["bot"].get("debug"):
        discord_logger.setLevel(logging.DEBUG)

    if not os.path.isdir('.logs'):
        os.makedirs('.logs')

    discord_handler = RotatingFileHandler(
        filename=".logs/discord.log",
        encoding="utf-8",
        mode="a",
        maxBytes=10 ** 7,
        backupCount=5,
    )

    if config["bot"].get("log_type") == "Timed":
        discord_handler = TimedRotatingFileHandler(
            filename=".logs/discord.log",
            when="midnight",
            interval=1,
            backupCount=5,
            encoding="utf-8",
        )

    discord_logger.addHandler(discord_handler)

    # Faster Event Loop
    try:
        import uvloop

        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except ImportError:
        pass

    # Initialize Bot
    global bot
    bot = ErinClient(config)
    slash = SlashCommand(bot, override_type=True, sync_commands=True)
    bot.remove_command("help")
    bot.setup()

    try:
        bot.run(config["bot"]["token"], bot=True, reconnect=True)
    except (KeyboardInterrupt, SystemExit):
        bot.close()
    finally:
        sys.exit()


if __name__ == "__main__":
    start()



from abc import ABC, abstractmethod


class DatabaseDriverBase(ABC):
    """
    Force implementation of some methods in the event a new driver is
    added to the database package.
    """

    @abstractmethod
    async def upsert(self, entity, states):
        pass

    @abstractmethod
    async def get(self, entity, state):
        pass

    @abstractmethod
    async def increment(self, entity, state, value):
        pass



from cookiecutter.main import cookiecutter

from erin.cli import CommandFactory


class ScaffoldCommand(CommandFactory):
    def __init__(self, parser, *args, **kwargs):
        self.parser = parser.add_parser(*args, **kwargs)
        self.parser.set_defaults(action=self.run)

    def run(self, *sys_args, **kwargs):
        print("Enter the details for your project below!")
        extra_context = {
            "project_name": ("Project Name", "Bot"),
            "project_slug": ("Project Slug", "bot"),
            "version": ("Version Number", "0.0.0.dev0"),
            "config_file": ("Configuration File Name", "config.toml"),
            "_copy_without_render": ["plugins/*"],
        }
        for key, val in extra_context.items():
            if isinstance(val, tuple):
                setting = input(f"{val[0]} [{val[1]}]: ")
                if len(setting.strip()) == 0:
                    extra_context[key] = val[1]
                else:
                    extra_context[key] = setting

        cookiecutter(
            "https://github.com/OpenDebates/cookiecutter-erin.git",
            no_input=True,
            extra_context=extra_context,
        )



import logging
import urllib.parse

from motor.motor_asyncio import AsyncIOMotorClient

from erin.db.abc import DatabaseDriverBase

logger = logging.getLogger(__name__)


class MongoClient(AsyncIOMotorClient, DatabaseDriverBase):
    def __init__(self, config, bot=None, *args, **kwargs):
        # Internal
        self.conn = None

        # Config
        self.database = config["database"]["database"]
        try:
            self.uri = config["database"]["uri"]
        except KeyError as e:
            self.uri = None
            self.host = config["database"]["host"]
            self.port = config["database"]["port"]
            self.username = urllib.parse.quote_plus(config["database"]["username"])
            self.password = urllib.parse.quote_plus(config["database"]["password"])
            self.replica_set = config["database"]["replica_set"]

        if self.uri:
            super(MongoClient, self).__init__(self.uri, *args, **kwargs)
            return

        # URI Building
        if len(self.host) == 1:
            self.uri = (
                f"mongodb://{self.username}:{self.password}"
                f"@{self.host[0]}:{self.port}"
            )
            super(MongoClient, self).__init__(self.uri, *args, **kwargs)
        elif len(self.host) > 1:
            host_list = []
            for replica_host in self.host:
                host = f"mongodb://{self.username}:{self.password}" f"@{replica_host}"
                host_list.append(host)

            super(MongoClient, self).__init__(
                host_list, replicaSet=f"{self.replica_set}", *args, **kwargs
            )

    async def upsert(self, entity, **states):
        """
        Updates an existing state's value. Creates a state
        if it does not exist. Also creates a database collection
        for each entity type when needed.

        :param entity: Any discord object with an id attribute
        :param states: A dict of state and possible values
        """
        if not (hasattr(entity, "id")):
            raise TypeError(f"'{entity}' is not an Entity!")

        collection = self[self.database][f"{entity.__class__.__name__}States"]
        await collection.update_many(
            {f"{entity.__class__.__name__.lower()}_id": entity.id},
            {"$set": states},
            upsert=True,
        )

    async def get(self, entity, state):
        """
        Grabs the value stored for an entity's state.

        :param entity: Any discord object with an id attribute
        :param state: An event passed as str
        :return: Returns the state's value if found or returns None
        """
        if not (hasattr(entity, "id")):
            raise TypeError(f"'{entity}' is not an Entity!")

        collection = self[self.database][f"{entity.__class__.__name__}States"]
        record = await collection.find_one(
            {f"{entity.__class__.__name__.lower()}_id": entity.id}
        )
        if record is None:
            return record
        else:
            try:
                state = record[state]
            except KeyError:
                return None
            return state

    async def increment(self, entity, state, value):
        """
        Increments an existing state's value.

        :param entity: Any discord object with an id attribute
        :param state: A state of type int
        :param value: The value to increment the state by
        """
        if not (hasattr(entity, "id")):
            raise TypeError(f"'{entity}' is not an Entity!")

        collection = self[self.database][f"{entity.__class__.__name__}States"]
        await collection.update_many(
            {f"{entity.__class__.__name__.lower()}_id": entity.id},
            {"$inc": {state: value}},
        )



import logging

# Access Logger
access_logger = logging.getLogger("access")
access_logger.setLevel(logging.INFO)


def get_hypercorn_logger(log_level: str) -> dict:
    log_level = log_level.upper()
    hypercorn_logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "root": {"level": f"{log_level}"},
        "loggers": {
            "hypercorn.error": {
                "level": f"{log_level}",
                "handlers": ["error_console"],
                "propagate": True,
                "qualname": "hypercorn.error",
            },
            "hypercorn.access": {
                "level": f"{log_level}",
                "handlers": ["console"],
                "propagate": True,
                "qualname": "hypercorn.access",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "generic",
                "stream": "ext://sys.stdout",
            },
            "error_console": {
                "class": "logging.StreamHandler",
                "formatter": "generic",
                "stream": "ext://sys.stderr",
            },
        },
        "formatters": {
            "generic": {
                "format": "%(asctime)s %(name)-18s %(levelname)-8s %(message)s",
                "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
                "class": "logging.Formatter",
            }
        },
    }
    return hypercorn_logging_config



from discord.ext import commands


class SchemaCommand(commands.Cog, name="Schema Command"):
    def __init__(self, bot):
        self.bot = bot

    @commands.command(name="schema")
    async def broken(self, ctx):
        await ctx.send("Sorry, I'm a broken schema!")


def setup(bot):
    bot.add_cog(SchemaCommand(bot))



import os

from schema import Optional, Or, Schema

ENV_MAPPINGS = {
    "bot": {
        "token": "ERIN_TOKEN",
        "debug": "ERIN_DEBUG",
        "project": "ERIN_PROJECT",
        "project_folder": "ERIN_PROJECT_FOLDER",
        "log_type": "ERIN_LOG_TYPE",
        "log_level": "ERIN_LOG_LEVEL",
    },
    "database": {
        "enabled": "ERIN_DB_ENABLED",
        "driver": "ERIN_DB_DRIVER",
        "uri": "ERIN_DB_URI",
        "host": "ERIN_DB_HOST",
        "port": "ERIN_DB_PORT",
        "username": "ERIN_DB_USERNAME",
        "password": "ERIN_DB_PASSWORD",
        "database": "ERIN_DB_DATABASE",
    },
    "global": {"prefixes": "ERIN_PREFIXES", "description": "ERIN_DESCRIPTION"},
}

OPTIONAL_ENVS = [
    "ERIN_DEBUG",
    "ERIN_DB_DRIVER",
    "ERIN_DB_URI",
    "ERIN_DB_HOST",
    "ERIN_DB_PORT",
    "ERIN_DB_USERNAME",
    "ERIN_DB_PASSWORD",
    "ERIN_DB_DATABASE" "ERIN_DESCRIPTION",
]


config_schema = Schema(
    {
        "bot": {
            "token": str,
            "debug": Or(True, False),
            "project": str,
            "plugins_folder": os.path.exists,
            "log_type": Or("Normal", "Timed"),
            "log_level": Or(
                "SPAM",
                "DEBUG",
                "VERBOSE",
                "INFO",
                "NOTICE",
                "WARNING",
                "SUCCESS",
                "ERROR",
                "CRITICAL",
            ),
        },
        "database": {
            # Schema hooks can be used to force driver detail checks
            # as noted in https://git.io/fhhd2 instead of resorting
            # to blanket optionals. Will get to this later if more
            # databases are needed!
            "enabled": Or(True, False),
            Optional("driver"): "mongo",
            Optional("uri"): str,
            Optional("host"): [str],
            Optional("port"): int,
            Optional("username"): str,
            Optional("password"): str,
            Optional("database"): str,
            Optional("replica"): str,
        },
        "global": {"name": str, "prefixes": [str], "description": str},
    },
    ignore_extra_keys=True,
)


plugin_schema = Schema({Optional("name"): str, "database": Or(True, False)})



from discord.ext import commands


class BrokenCommand(commands.Cog, name="Broken Command"):
    def __init__(self, bot):
        self.bot = bot

    @commands.command(name="broken")
    async def broken(self, ctx):
        await ctx.send("Sorry, I'm broken!")


def setup(bot):
    bot.add_cog(BrokenCommand(bot))



import logging

import discord
from discord.ext import commands
from schema import SchemaError

from erin.core.schema import plugin_schema
from erin.core.utils import find_plugins, get_plugin_data
from erin.db.drivers import MongoClient

# Logging
logger = logging.getLogger(__name__)


class ErinClient(commands.Bot):
    """
    Custom implementation designed to load configuration from the TOML
    config file and dynamic console configurations
    """

    def __init__(self, config, *args, **kwargs):
        self.config = config

        super().__init__(
            command_prefix=self._get_command_prefix(),
            description=self._get_description(),
            intents=discord.Intents.all(),
            *args,
            **kwargs,
        )

        # Logger
        self.logger = logger

        # Database
        if config["database"].get("enabled"):
            self.db = MongoClient(config, bot=self)
        else:
            self.logger.notice("No database defined. Running without one!")

    def _get_command_prefix(self):
        self.prefixes = self.config["global"]["prefixes"]
        return self.prefixes

    def _get_description(self):
        self.description = self.config["global"]["description"]
        if self.description:
            return self.description
        else:
            return ""

    def _load_plugins(self):
        plugin_dir = self.config["bot"].get("plugins_folder")
        try:
            extensions = find_plugins(plugin_dir)
            logger.verbose(f"Plugins Found: {extensions}")
        except Exception:
            self.logger.exception()
            self.logout()
            return None
        for extension in extensions:

            plugin_data = get_plugin_data(extension)
            if not plugin_data:
                self.logger.notice(f"Skipping {extension}: `plugin_data` undefined")
                continue

            # Convert to database method later
            if not self.config["database"].get("enabled"):
                if plugin_data.get("database"):
                    logger.notice(f"Skipping {extension}: Database Needed")
                    continue
            else:
                try:
                    plugin_schema.validate(plugin_data)
                except SchemaError as e:
                    self.logger.exception(f"Plugin data is invalid: {extension}")

            if plugin_data.get("name"):
                self.logger.verbose(f"Loading Plugin: {plugin_data['name']}")
            else:
                self.logger.verbose(f"Loading Plugin: {extension}")

            # Attempt loading the plugin
            try:
                self.load_extension(extension)
            except discord.ClientException:
                self.logger.exception(f"Missing setup() for Plugin: {extension}.")
            except ImportError:
                self.logger.exception(f"Failed to load Plugin: {extension}")
            except Exception as e:
                self.logger.exception("Core Error:")

    def setup(self):
        """
        Important setup functions and their configurations have to be
        called here.
        """
        self._load_plugins()

    async def on_command(self, ctx):
        if ctx.cog.__cog_name__:
            cog_name = ctx.cog.__cog_name__
        else:
            cog_name = repr(ctx.cog)

        self.logger.info(
            f"Cog: {cog_name} "
            f"| Invoked With: {ctx.invoked_with} "
            f"| Invoked By: {ctx.message.author} - <@!{ctx.message.author.id}> "
            f"| Message Content: \n"
            f"{ctx.message.content}"
        )
